#include <FastLED.h>
#include <LEDMatrix.h>
#include <LEDSprites.h>
#include <LEDText.h>
#include <FontMatrise.h>
#include "BluetoothSerial.h"

#define LED_PIN        2
#define COLOR_ORDER    GRB
#define CHIPSET        WS2812B
#define MATRIX_WIDTH   16
#define MATRIX_HEIGHT  16
#define MATRIX_TYPE    VERTICAL_ZIGZAG_MATRIX

#define APP_MENU -1
#define APP_TETRIS 0
#define APP_SNAKE 1
#define APP_BREAKOUT 2
#define APP_MEMGAME 3
#define APP_PIXEL 4
#define APP_SENSOR 5

int currentApp = APP_MENU;

cLEDMatrix<MATRIX_WIDTH, -MATRIX_HEIGHT, MATRIX_TYPE> leds;
BluetoothSerial SerialBT;

// Bluetooth input
enum btnInput {NONE, UP, DOWN, LEFT, RIGHT};
btnInput currentInput = NONE;

#include "Tetris.h"
#include "Menu.h"
#include "Snake.h"
#include "Breakout.h"
#include "Pixel.h"
#include "Sensor.h"
#include "MemGame.h"

void setup()
{
  Serial.begin(115200);
  SerialBT.begin("ESP32Matrix");
  
  FastLED.addLeds<CHIPSET, LED_PIN, COLOR_ORDER>(leds[0], leds.Size());
  // FastLED.setMaxPowerInVoltsAndMilliamps(5,2000);
  FastLED.clear(true);
  FastLED.show();
}

void loop()
{
  switch (currentApp) {
    case APP_MENU : runMenu();
    break;
    case APP_TETRIS : runTetris();
    break;
    case APP_SNAKE : runSnake();
    break;
    case APP_BREAKOUT : runBreakout();
    break;
    case APP_MEMGAME: runMemGame();
    break;
    case APP_PIXEL : runPixel();
    break;
    case APP_SENSOR: runSensor();
    break;
  }
}

void runMenu(){
  bool isRunning = true;
  Menu menu = Menu();
  menu.setup();
  while (isRunning) {
    isRunning = menu.loop();
  }
}

void runTetris(){
  bool isRunning = true;
  Tetris tetris = Tetris();
  tetris.setup();
  while (isRunning) {
    isRunning = tetris.loop();
  }
}

void runSnake(){
  bool isRunning = true;
  Snake snake = Snake();
  snake.setup();
  while (isRunning) {
    isRunning = snake.loop();
  }
}

void runBreakout(){
  bool isRunning = true;
  Breakout breakout = Breakout();
  breakout.setup();
  while (isRunning) {
    isRunning = breakout.loop();
  }
}

void runPixel(){
  bool isRunning = true;
  Pixel pixel = Pixel();
  pixel.setup();
  while (isRunning) {
    isRunning = pixel.loop();
  }
}

void runSensor(){
  bool isRunning = true;
  Sensor sensor = Sensor();
  sensor.setup();
  while (isRunning) {
    isRunning = sensor.loop();
  }
}

void runMemGame(){
  bool isRunning = true;
  MemGame memgame = MemGame();
  memgame.setup();
  while (isRunning) {
    isRunning = memgame.loop();
  }
}
