#include <FastLED.h>
#include <LEDText.h>
#include <TinyFont.h>

enum Colors { BLUE, RED, GREEN, YELLOW, BADINPUT };

class MemGame {
  #define BRIGHT 255
  #define DIM 64

  int ledBrightnessBlue, ledBrightnessRed, ledBrightnessGreen, ledBrightnessYellow;
  int gameLength = 0;
  Colors gameSequence[100];
  bool turn;
  int playerIdx;
  bool gameOver;
  

  cLEDText ScrollingMsg;
  unsigned char txtMemGame[16];
  uint16_t PlasmaTime, PlasmaShift;
  const uint8_t PlasmaLowHue = 100;
  const uint8_t PlasmaHighHue = 150;

public:
  MemGame() {
  }

  void setup() {
    sprintf((char *)txtMemGame, "                ");
    ScrollingMsg.SetFont(MatriseFontData);
    ScrollingMsg.Init(&leds, leds.Width(), ScrollingMsg.FontHeight() + 1, 0, 5);
    ScrollingMsg.SetBackgroundMode(BACKGND_ERASE);
    ScrollingMsg.SetScrollDirection(SCROLL_LEFT);
    ScrollingMsg.SetFrameRate(1);
    ScrollingMsg.SetTextColrOptions(COLR_RGB | COLR_SINGLE, 0xFF, 0xFF, 0xFF);

    ledBrightnessRed = DIM;
    ledBrightnessBlue = DIM;
    ledBrightnessGreen = DIM;
    ledBrightnessYellow = DIM;
    gameSequence[gameLength] = static_cast<Colors>(random(4));
    turn = true;
    gameOver = false;
  }

  boolean loop() {

    if(gameOver) {
      if (ScrollingMsg.UpdateText() == -1) {
        sprintf((char *)txtMemGame, "   SCORE %u", gameLength);
        ScrollingMsg.SetText((unsigned char *)txtMemGame, sizeof(txtMemGame)-1);
      }

      if(SerialBT.available()) {
          char keyPress = (char)SerialBT.read();
          if (keyPress) {
            switch (keyPress) {
              case 'm': currentApp = -1; return false;
              default: return false;
            }
          }
      }

      FastLED.show();
      delay(4);
    }
    if(!gameOver) {
      if(turn) {

        for(int k = 0; k <= gameLength; k++) {
          switch(gameSequence[k]) {
            case BLUE: ledBrightnessBlue = BRIGHT;
            break;
            case GREEN: ledBrightnessGreen = BRIGHT;
            break;
            case YELLOW: ledBrightnessYellow = BRIGHT;
            break;
            case RED: ledBrightnessRed = BRIGHT;
            break;
          }

          for (int i = 0; i < 18; i++) {
            for (int j = 0; j < 18; j++) {
              if (i >= 8 && j < 8)
                leds(j, i) = CHSV(160, 255, ledBrightnessBlue);
              if (i < 8 && j < 8)
                leds(j, i) = CHSV(81, 255, ledBrightnessGreen);
              if (i < 8 && j >= 8)
                leds(j, i) = CHSV(43, 255, ledBrightnessYellow);
              if (i >= 8 && j >= 8)
                leds(j, i) = CHSV(0, 255, ledBrightnessRed);
            }
          }
          FastLED.show();

          delay(500);
          
          ledBrightnessRed = DIM;
          ledBrightnessBlue = DIM;
          ledBrightnessGreen = DIM;
          ledBrightnessYellow = DIM;

          for (int i = 0; i < 18; i++) {
            for (int j = 0; j < 18; j++) {
              if (i >= 8 && j < 8)
                leds(j, i) = CHSV(160, 255, ledBrightnessBlue);
              if (i < 8 && j < 8)
                leds(j, i) = CHSV(81, 255, ledBrightnessGreen);
              if (i < 8 && j >= 8)
                leds(j, i) = CHSV(43, 255, ledBrightnessYellow);
              if (i >= 8 && j >= 8)
                leds(j, i) = CHSV(0, 255, ledBrightnessRed);
            }
          }
          FastLED.show();
          delay(500);
        }
        turn = false;
        playerIdx = 0;

      } else {
        if(SerialBT.available()) {
          char keyPress = (char)SerialBT.read();
          if (keyPress) {
            Colors expectedColor;
            switch (keyPress) {
              case 'b': expectedColor = BLUE; break;
              case 'r': expectedColor = RED; break;
              case 'g': expectedColor = GREEN; break;
              case 'y': expectedColor = YELLOW; break;
              case 'm': currentApp = -1; return false;
              default: expectedColor = BADINPUT;
            }
            if (expectedColor != BADINPUT) {
              if(gameSequence[playerIdx] == expectedColor) {
                playerIdx++;
                if (playerIdx >= gameLength + 1) {
                  gameLength++;
                  gameSequence[gameLength] = static_cast<Colors>(random(4));
                  turn = true;
                }
              } else {
                // Incorrect input, game over
                gameOver = true;
              }
            }
          }
        }
      }
    }

    return true;
  }
};
