#include <FastLED.h>
#include <LEDText.h>
#include <TinyFont.h>
#include <DHT11.h>

#define DHT11_PIN 4
DHT11 dht11(DHT11_PIN);

class Sensor {
  cLEDText TempMsg, HumMsg;
  uint16_t PlasmaTime, PlasmaShift;
  const uint8_t PlasmaLowHue = 100;
  const uint8_t PlasmaHighHue = 150;
  char temp_string[12];
  char hum_string[12];
  unsigned long lastReadTempTime = 0;
  int temperature, humidity;

public:
  Sensor() {
  }

  void setup() {
    TempMsg.SetFont(TinyFontData);
    TempMsg.Init(&leds, 16, TempMsg.FontHeight() + 1, 2, 9);
    TempMsg.SetBackgroundMode(BACKGND_LEAVE);
    TempMsg.SetFrameRate(5);
    TempMsg.SetTextColrOptions(COLR_RGB | COLR_SINGLE, 0xFF, 0xFF, 0xFF);
    TempMsg.SetText((unsigned char *)"0123", 5);

    HumMsg.SetFont(TinyFontData);
    HumMsg.Init(&leds, 16, TempMsg.FontHeight() + 1, 2, 2);
    HumMsg.SetBackgroundMode(BACKGND_LEAVE);
    HumMsg.SetFrameRate(5);
    HumMsg.SetTextColrOptions(COLR_RGB | COLR_SINGLE, 0xFF, 0xFF, 0xFF);
    HumMsg.SetText((unsigned char *)"4567", 5);
  }

  boolean loop() {
    // Fill background with dim plasma
    #define PLASMA_X_FACTOR  48
    #define PLASMA_Y_FACTOR  48
    for (int16_t x = 0; x < leds.Width(); x++) {
      for (int16_t y = 0; y < leds.Height(); y++) {
        int16_t r = sin16(PlasmaTime) / 256;
        int16_t h = sin16(x * r * PLASMA_X_FACTOR + PlasmaTime) + cos16(y * (-r) * PLASMA_Y_FACTOR + PlasmaTime) + sin16(y * x * (cos16(-PlasmaTime) / 256) / 2);
        int16_t hue = ((h / 256) + 128);
        leds(x, y) = CHSV((uint8_t)hue, 255, 64);
      }
    }
    uint16_t OldPlasmaTime = PlasmaTime;
    PlasmaTime += PlasmaShift;
    if (OldPlasmaTime > PlasmaTime) {
      PlasmaShift = (random8(0, 5) * 32) + 64;
    }

    if(millis() - lastReadTempTime >= 5000) {
      // Read temperature and humidity
      dht11.readTemperatureHumidity(temperature, humidity);
      sprintf(temp_string, "%dC", temperature);
      sprintf(hum_string, "%dB", humidity);

      lastReadTempTime = millis();
    }
    TempMsg.UpdateText();
    TempMsg.SetText((unsigned char *)temp_string, strlen((const char *)temp_string));

    HumMsg.UpdateText();
    HumMsg.SetText((unsigned char *)hum_string, strlen((const char *)hum_string));

    FastLED.show();
    if(SerialBT.available()){
        char keyPress = (char)SerialBT.read();
        switch(keyPress) {
          case 'w':
          case 'a':
          case 's':
          case 'd':
          case 'm':
            currentApp = -1;
            return false;
        }
      }

    return true;
  }
};
